/* mbsStatusRecord.c */
/* Example record support module */
  
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "alarm.h"
#include "dbAccess.h"
#include "recGbl.h"
#include "dbEvent.h"
#include "dbDefs.h"
#include "dbAccess.h"
#include "devSup.h"
#include "errMdef.h"
#include "recSup.h"
#include "special.h"
#define GEN_SIZE_OFFSET
#include "mbsStatusRecord.h"
#undef  GEN_SIZE_OFFSET
#include "epicsExport.h"

/* Create RSET - Record Support Entry Table */
#define report NULL
#define initialize NULL
static long init_record();
static long process();
#define special NULL
#define get_value NULL
#define cvt_dbaddr NULL
#define get_array_info NULL
#define put_array_info NULL
static long get_units();
static long get_precision();
#define get_enum_str NULL
#define get_enum_strs NULL
#define put_enum_str NULL
static long get_graphic_double();
static long get_control_double();
static long get_alarm_double();
 
rset mbsStatusRSET={
	RSETNUMBER,
	report,
	initialize,
	init_record,
	process,
	special,
	get_value,
	cvt_dbaddr,
	get_array_info,
	put_array_info,
	get_units,
	get_precision,
	get_enum_str,
	get_enum_strs,
	put_enum_str,
	get_graphic_double,
	get_control_double,
	get_alarm_double
};
epicsExportAddress(rset,mbsStatusRSET);

typedef struct mbsStatusset { /* mbsStatus input dset */
	long		number;
	DEVSUPFUN	dev_report;
	DEVSUPFUN	init;
	DEVSUPFUN	init_record; /*returns: (-1,0)=>(failure,success)*/
	DEVSUPFUN	get_ioint_info;
	DEVSUPFUN	read_mbsStatus;
}mbsStatusdset;

static void checkAlarms(mbsStatusRecord *pmbsStatus);
static void monitor(mbsStatusRecord *pmbsStatus);

static long init_record(void *precord,int pass)
{
    mbsStatusRecord	*pmbsStatus = (mbsStatusRecord *)precord;
    mbsStatusdset	*pdset;
    long	status;

    if (pass==0) return(0);

    if(!(pdset = (mbsStatusdset *)(pmbsStatus->dset))) {
	recGblRecordError(S_dev_noDSET,(void *)pmbsStatus,"mbsStatus: init_record");
	return(S_dev_noDSET);
    }
    /* must have read_mbsStatus function defined */
    if( (pdset->number < 5) || (pdset->read_mbsStatus == NULL) ) {
	recGblRecordError(S_dev_missingSup,(void *)pmbsStatus,"mbsStatus: init_record");
	return(S_dev_missingSup);
    }

    if( pdset->init_record ) {
	if((status=(*pdset->init_record)(pmbsStatus))) return(status);
    }
    return(0);
}

static long process(void *precord)
{
	mbsStatusRecord	*pmbsStatus = (mbsStatusRecord *)precord;
	mbsStatusdset		*pdset = (mbsStatusdset *)(pmbsStatus->dset);
	long		 status;
	unsigned char    pact=pmbsStatus->pact;

	if( (pdset==NULL) || (pdset->read_mbsStatus==NULL) ) {
		pmbsStatus->pact=TRUE;
		recGblRecordError(S_dev_missingSup,(void *)pmbsStatus,"read_mbsStatus");
		return(S_dev_missingSup);
	}

	/* pact must not be set until after calling device support */
	status=(*pdset->read_mbsStatus)(pmbsStatus);
	/* check if device support set pact */
	if ( !pact && pmbsStatus->pact ) return(0);
	pmbsStatus->pact = TRUE;

	recGblGetTimeStamp(pmbsStatus);
	/* check for alarms */
	checkAlarms(pmbsStatus);
	/* check event list */
	monitor(pmbsStatus);
	/* process the forward scan link record */
        recGblFwdLink(pmbsStatus);

	pmbsStatus->pact=FALSE;
	return(status);
}

static long get_units(DBADDR *paddr, char *units)
{
    mbsStatusRecord	*pmbsStatus=(mbsStatusRecord *)paddr->precord;

    strncpy(units,pmbsStatus->egu,DB_UNITS_SIZE);
    return(0);
}

static long get_precision(DBADDR *paddr, long *precision)
{
    mbsStatusRecord	*pmbsStatus=(mbsStatusRecord *)paddr->precord;

    *precision = pmbsStatus->prec;
    if(paddr->pfield == (void *)&pmbsStatus->val) return(0);
    recGblGetPrec(paddr,precision);
    return(0);
}

static long get_graphic_double(DBADDR *paddr,struct dbr_grDouble *pgd)
{
    mbsStatusRecord	*pmbsStatus=(mbsStatusRecord *)paddr->precord;
    int		fieldIndex = dbGetFieldIndex(paddr);

    if(fieldIndex == mbsStatusRecordVAL
    || fieldIndex == mbsStatusRecordHIHI
    || fieldIndex == mbsStatusRecordHIGH
    || fieldIndex == mbsStatusRecordLOW
    || fieldIndex == mbsStatusRecordLOLO
    || fieldIndex == mbsStatusRecordHOPR
    || fieldIndex == mbsStatusRecordLOPR) {
        pgd->upper_disp_limit = pmbsStatus->hopr;
        pgd->lower_disp_limit = pmbsStatus->lopr;
    } else recGblGetGraphicDouble(paddr,pgd);
    return(0);
}

static long get_control_double(DBADDR *paddr,struct dbr_ctrlDouble *pcd)
{
    mbsStatusRecord	*pmbsStatus=(mbsStatusRecord *)paddr->precord;
    int		fieldIndex = dbGetFieldIndex(paddr);

    if(fieldIndex == mbsStatusRecordVAL
    || fieldIndex == mbsStatusRecordHIHI
    || fieldIndex == mbsStatusRecordHIGH
    || fieldIndex == mbsStatusRecordLOW
    || fieldIndex == mbsStatusRecordLOLO) {
	pcd->upper_ctrl_limit = pmbsStatus->hopr;
	pcd->lower_ctrl_limit = pmbsStatus->lopr;
    } else recGblGetControlDouble(paddr,pcd);
    return(0);
}

static long get_alarm_double(DBADDR *paddr,struct dbr_alDouble *pad)
{
    mbsStatusRecord	*pmbsStatus=(mbsStatusRecord *)paddr->precord;
    int		fieldIndex = dbGetFieldIndex(paddr);

    if(fieldIndex == mbsStatusRecordVAL) {
	pad->upper_alarm_limit = pmbsStatus->hihi;
	pad->upper_warning_limit = pmbsStatus->high;
	pad->lower_warning_limit = pmbsStatus->low;
	pad->lower_alarm_limit = pmbsStatus->lolo;
    } else recGblGetAlarmDouble(paddr,pad);
    return(0);
}

static void checkAlarms(mbsStatusRecord *pmbsStatus)
{
	double		val;
	float		hyst, lalm, hihi, high, low, lolo;
	unsigned short	hhsv, llsv, hsv, lsv;

	if(pmbsStatus->udf == TRUE ){
		recGblSetSevr(pmbsStatus,UDF_ALARM,INVALID_ALARM);
		return;
	}
	hihi = pmbsStatus->hihi; lolo = pmbsStatus->lolo; high = pmbsStatus->high; low = pmbsStatus->low;
	hhsv = pmbsStatus->hhsv; llsv = pmbsStatus->llsv; hsv = pmbsStatus->hsv; lsv = pmbsStatus->lsv;
	val = pmbsStatus->val; hyst = pmbsStatus->hyst; lalm = pmbsStatus->lalm;

	/* alarm condition hihi */
	if (hhsv && (val >= hihi || ((lalm==hihi) && (val >= hihi-hyst)))){
	        if (recGblSetSevr(pmbsStatus,HIHI_ALARM,pmbsStatus->hhsv)) pmbsStatus->lalm = hihi;
		return;
	}

	/* alarm condition lolo */
	if (llsv && (val <= lolo || ((lalm==lolo) && (val <= lolo+hyst)))){
	        if (recGblSetSevr(pmbsStatus,LOLO_ALARM,pmbsStatus->llsv)) pmbsStatus->lalm = lolo;
		return;
	}

	/* alarm condition high */
	if (hsv && (val >= high || ((lalm==high) && (val >= high-hyst)))){
	        if (recGblSetSevr(pmbsStatus,HIGH_ALARM,pmbsStatus->hsv)) pmbsStatus->lalm = high;
		return;
	}

	/* alarm condition low */
	if (lsv && (val <= low || ((lalm==low) && (val <= low+hyst)))){
	        if (recGblSetSevr(pmbsStatus,LOW_ALARM,pmbsStatus->lsv)) pmbsStatus->lalm = low;
		return;
	}

	/* we get here only if val is out of alarm by at least hyst */
	pmbsStatus->lalm = val;
	return;
}

static void monitor(mbsStatusRecord *pmbsStatus)
{
  unsigned short	monitor_mask;
  double		delta;

  monitor_mask = recGblResetAlarms(pmbsStatus);
  /* check for value change */
  delta = pmbsStatus->mlst - pmbsStatus->val;
  if(delta<0.0) delta = -delta;
  if (delta > pmbsStatus->mdel) {
    /* post events for value change */
    monitor_mask |= DBE_VALUE;
    /* update last value monitored */
    pmbsStatus->mlst = pmbsStatus->val;
  }

  /* check for archive change */
  delta = pmbsStatus->alst - pmbsStatus->val;
  if(delta<0.0) delta = -delta;
  if (delta > pmbsStatus->adel) {
    /* post events on value field for archive change */
    monitor_mask |= DBE_LOG;
    /* update last archive value monitored */
    pmbsStatus->alst = pmbsStatus->val;
  }

  /* send out monitors connected to the value field */
  if (monitor_mask){
    db_post_events(pmbsStatus,&pmbsStatus->val,monitor_mask);
  }
  return;
}
